#!/usr/bin/env python
"""
A python module that uses pyepics to save/restore sets of pvs from files.

Copyright 2011 Angus Gratton <angus.gratton@anu.edu.au>
Australian National University
EPICS Open License

The module is intended to be compatible with the 'autosave' module format used in synApps.

Files -

xxx.req - A request file with a list of pvs to save. Format is the same as autosave request format,
          including being able to have "file yyy.req VAR=A,OTHER=B" style macro expansions.

xxx.sav - A saved file with the current PV values, to save/restore. Standalone file, this is a
          compatible format to the .sav files which are used by autosave.

This module requires/uses pyparsing parser framework. Debian/Ubuntu package is "python-pyparsing"
Web site is http://pyparsing.wikispaces.com/

"""

from pyparsing import (Literal, Optional, Word, Combine, Regex, Group,
                       ZeroOrMore, OneOrMore, LineEnd, LineStart, StringEnd,
                       alphanums, alphas, nums, printables)

import sys
import os
import datetime
import json
import time
from epics.pv import get_pv
from epics.utils import IOENCODING


def restore_pvs(filepath, debug=False):
    """
    Restore pvs from a save file via Channel Access

    debug - Set to True if you want a line printed for each value set

    Returns True if all pvs were restored successfully.
    """
    pv_vals = []
    failures = []
    # preload PV names and values, hoping PV connections happen in background
    with open(filepath, 'r', encoding=IOENCODING) as fh:
        for line in fh.readlines():
            if (len(line) < 2 or line.startswith('<END') or line.startswith('#')
                or len(line[:-1].split(' ', 1))<2):
                continue
            pvname, value = [w.strip() for w in line[:-1].split(' ', 1)]
            if value.startswith('@array@'):
                value = value.replace('@array@', '').strip()
                if value.startswith('{') and value.endswith('}'):
                    value = value[1:-1]
                value = json.loads(value)
            thispv = get_pv(pvname, connect=False)
            pv_vals.append((thispv, value))

    for thispv, value in pv_vals:
        thispv.connect()
        pvname = thispv.pvname
        if not thispv.connected:
            time.sleep(0.1)
        if not thispv.connected:
            print("Cannot connect to %s" % (pvname))
        elif not thispv.write_access:
            print("No write access to %s" % (pvname))
        else:
            if debug:
                print("Setting %s to %s" % (pvname, value))
            try:
                thispv.put(value, wait=False)
            except:
                exctype, excvalue, exctrace = sys.exc_info()
                print("Error restoring %s to %s : %s" % (pvname, value,
                                                         exctype, excvalue))
                failues.append(pvname)
    return len(failures) == 0

def save_pvs(request_file, save_file, debug=False):
    """
    Save pvs from a request file to a save file, via Channel Access

    Set debug=True to print a line for each PV saved.

    Will print a warning if a PV cannot connect.
    """
    saver = AutoSaver(request_file)
    saver.save(save_file, verbose=debug)

class AutoSaver(object):
    """Autosave class"""
    def __init__(self, request_file=None):
        self.request_file = request_file
        self.pvs = []
        if request_file is not None:
            self.read_request_file(request_file)

    def read_request_file(self, request_file=None):
        if request_file is not None:
            self.request_file = request_file
        self.pvs = []
        for pvname in _parse_request_file(request_file):
            self.pvs.append(get_pv(pvname, connect=False))

    def save(self, save_file=None, verbose=False):
        """save PVs to save_file"""
        now = datetime.datetime.now()
        if save_file is None:
            sfile = self.request_file
            if sfile.endswith('.req'):
                sfile = sfile[:-4]
            tstamp = now.strftime("%Y%b%d_%H%M%S")
            save_file = "%s_%s.sav" % (sfile, tstamp)

        buff = ["# File saved by pyepics AutoSaver.save() on %s" % now,
                "# Edit with extreme care."]

        for thispv in self.pvs:
            pvname = thispv.pvname
            thispv.wait_for_connection()
            if thispv.connected:
                if thispv.count == 1:
                    value = str(thispv.get())
                elif thispv.count > 1 and 'char' in thispv.type:
                    value = thispv.get(as_string=True)
                elif thispv.count > 1 and 'char' not in thispv.type:
                    value = '@array@ %s' % json.dumps(thispv.get().tolist())
                buff.append("%s %s" % (pvname, value))
                if verbose:
                    print( "PV %s = %s" % (pvname, value))
            elif verbose:
                print("PV %s not connected" % (pvname))


        buff.append("<END>\n")
        with open(save_file, 'w', encoding=IOENCODING) as fh:
            fh.write("\n".join(buff))
        print("wrote %s"% save_file)

def _parse_request_file(request_file, macro_values={}):
    """
    Internal function to parse a request file.

    Parse happens in two stages, first build an AST then walk it and do
    file expansions (which recurse through here.)

    Returns a list of PV names.

    """
    ast = [ x for x in req_file.parseFile(request_file).asList() if len(x) > 0 ]

    result = []
    for n in ast:
        if len(n) == 1: # simple PV name
            pvname = n[0]
            for m,v in macro_values.items(): # please forgive me this awful macro expansion method
                pvname = pvname.replace("$(%s)" % m, v)
            result.append(pvname)
        elif n[0] == 'file': # include file
            subfile = n[1]
            subfile = os.path.normpath(os.path.join(os.path.dirname(request_file), subfile))
            sub_macro_vals = macro_values.copy()

            for m, v in n[2:]:
                if v != '$({})'.format(m):
                    #Allows propagation of macro values in the usual way
                    sub_macro_vals[m] = v
            result += _parse_request_file(subfile, sub_macro_vals)
        else:
            raise Exception("Unexpected entry parsed from request file: %s" % n)
    return result

# request & save file grammar (combined because lots of it is pretty similar)
point = Literal('.')
minus = Literal('-')
ignored_quote = Literal('"').suppress()
ignored_comma = Literal(',').suppress()

file_name = Word(alphanums+":._-+/\\")

number = Word(nums)
integer = Combine( Optional(minus) + number )
float_number = Combine( integer +
                        Optional( point + Optional(number) )
                        ).setParseAction(lambda t:float(t[0]))

# PV names according to app developer guide and tech-talk email thread at:
# https://epics.anl.gov/tech-talk/2019/msg01429.php
pv_name = Combine(Word(alphanums+'$()_-+:[]<>;{}')
                  + Optional(Combine('.') + Word(printables)))
pv_value = (float_number | Word(printables))

pv_assignment = pv_name + pv_value

comment = Literal("#") + Regex(r".*")

macro = Group( Word(alphas) + Literal("=").suppress() + pv_name )
macros = Optional(macro + ZeroOrMore(Word(";,").suppress() + macro) )

#file_include = Literal("file") + pv_name + macros
file_include = Literal("file") + \
               (file_name | ignored_quote + file_name + ignored_quote) \
               + Optional(ignored_comma) + macros

def line(contents):
    return LineStart() + ZeroOrMore(Group(contents)) + LineEnd().suppress()

req_line = line( file_include | comment.suppress() | pv_name )
req_file = OneOrMore(req_line) + StringEnd().suppress()

sav_line = line( comment.suppress() | Literal("<END>").suppress() | pv_assignment)
sav_file = OneOrMore(sav_line) + StringEnd().suppress()



# -*- coding: utf-8 -*-
import sys, os
from packaging.version import parse as version_parse
import epics

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.mathjax',
              'sphinx.ext.napoleon', 'sphinxcontrib.video',
              'sphinx_copybutton', 'numpydoc']
templates_path = ['_templates']


project = "PyEpics"
copyright = "2025, Matthew Newville, The University of Chicago"

html_title = "Epics Channel Access for Python"
html_short_title = "PyEpics"

release = version_parse(epics.__version__).base_version

source_suffix = {'.rst': 'restructuredtext'}
exclude_trees = ['_build']
default_role = None
source_encoding = 'utf-8'

add_function_parentheses = True
add_module_names = True
pygments_style = 'sphinx'
master_doc = 'index'

html_theme_path = ['sphinx_theme']
html_theme = 'bizstyle'

html_static_path = ['_static']
html_sidebars = {
  'index': ["indexsidebar.html",  "sourcelink.html", "searchbox.html"],
  "**": [ "localtoc.html",  "relations.html", "sourcelink.html", "searchbox.html"]
}

html_domain_indices = False
html_use_index = True
html_show_sourcelink = True



#!/usr/bin/python
#
# simple PV Probe application for Float PVs

import wx
import sys
import epics
from epics.wx import PVText, PVFloatCtrl, PVFloatSpin

class ProbeFrame(wx.Frame):
    def __init__(self, parent=None, **kwds):

        wx.Frame.__init__(self, parent, wx.ID_ANY, '',
                         wx.DefaultPosition,  size=(400, 400), **kwds)
        self.SetTitle("Connect to Float PV:")

        self.SetFont(wx.Font(12,wx.SWISS,wx.NORMAL,wx.BOLD,False))

        sizer = wx.GridBagSizer(3, 3)
        panel = wx.Panel(self)

        self.pvname = wx.TextCtrl(panel, value='', size=(150, -1),
                                      style=wx.TE_PROCESS_ENTER)
        self.pvname.Bind(wx.EVT_CHAR, self.onPVName)

        self.pvtext = PVText(panel, None, size=(100, -1))
        self.pvfloat = PVFloatCtrl(panel, None, size=(100, -1))
        self.pvfspin = PVFloatSpin(panel, None, size=(100, -1))

        pvname_label = wx.StaticText(panel, label='PV Name:', size=(150, -1))
        pvval_label = wx.StaticText(panel, label='PVText:', size=(150, -1))
        pvfloat_label = wx.StaticText(panel, label='PVFloatControl:', size=(150, -1))
        pvfspin_label = wx.StaticText(panel, label='PVFloatSpin:', size=(150, -1))

        sizer.Add(pvname_label,  (0, 0), (1, 1), wx.ALIGN_LEFT, 1)
        sizer.Add(self.pvname,   (0, 1), (1, 1), wx.ALIGN_LEFT, 1)
        sizer.Add(pvval_label,   (1, 0), (1, 1), wx.ALIGN_LEFT, 1)
        sizer.Add(self.pvtext,   (1, 1), (1, 1), wx.ALIGN_LEFT, 1)
        sizer.Add(pvfloat_label,  (2, 0), (1, 1), wx.ALIGN_LEFT, 1)
        sizer.Add(self.pvfloat,   (2, 1), (1, 1), wx.ALIGN_LEFT, 1)
        sizer.Add(pvfspin_label,  (3, 0), (1, 1), wx.ALIGN_LEFT, 1)
        sizer.Add(self.pvfspin,   (3, 1), (1, 1), wx.ALIGN_LEFT, 1)

        panel.SetSizer(sizer)
        sizer.Fit(panel)

        s = wx.BoxSizer(wx.VERTICAL)
        s.Add(panel)
        s.Fit(self)

        self.SetSize((400, 150))
        self.Refresh()

    def onPVName(self, event=None):
        if event.GetKeyCode() == wx.WXK_RETURN:
            pvname = self.pvname.GetValue().strip()
            if len(pvname) > 1:
                self.pvtext.SetPV(pvname)
                self.pvfloat.SetPV(pvname)
                self.pvfspin.SetPV(pvname)
        event.Skip()

if __name__ == '__main__':
    app = wx.App(redirect=False)
    ProbeFrame().Show()
    app.MainLoop()



import time
from . import ca, dbr, pv, alarm, device, motor, multiproc
from .version import __version__

__doc__ = f"""
   Epics Channel Access Python module

   version: {__version__}
   Principal Authors:
      Matthew Newville <newville@cars.uchicago.edu> CARS, University of Chicago
      Ken Lauer, David Chabot, Angus Gratton

== License:
   Except where explicitly noted, this file and all files in this distribution
   are licensed under the Epics Open License See LICENSE in the top-level
   directory of this distribution.

== Overview:
   Python Interface to the Epics Channel Access protocol of the Epics control system.
"""

PV = pv.PV
Alarm = alarm.Alarm
Motor = motor.Motor
Device = device.Device
poll = ca.poll

get_pv = pv.get_pv

CAProcess = multiproc.CAProcess
CAPool = multiproc.CAPool

# some constants
NO_ALARM = 0
MINOR_ALARM = 1
MAJOR_ALARM = 2
INVALID_ALARM = 3

_PVmonitors_ = {}

def caput(pvname, value, wait=False, timeout=60.0, connection_timeout=5.0):
    """
    put a value to an epics Process Variable (PV).

    Parameters
    ----------
     pvname : str
         name of PV
     value : object
         value to put to PV (see notes)
     wait : bool, optional
         whether to wait for processing to complete [False]
     timeout : float, optional
         maximum time (in seconds) to wait for the processing to complete [60]
     connection_timeout : float, optional
         maximum time (in seconds) to wait for connection [5]

    Returns
    -------
    int or ``None``
      - 1  on succesful completion
      - -1 or other negative value on failure of the low-level CA put.
      - None  on a failure to connect to the PV.

    Notes
    -----
    1. Epics PVs are typically limited to an appropriate Epics data type,
       int, float, str, and enums or homogeneously typed lists or arrays.
       Numpy arrays or Python strings are generally coeced as appropriate,
       but care may be needed when mapping Python objects to Epics PV values.
    2. If not already connected, the PV will first attempt to connect to
       the networked variable. As the initial connection can take some time
       (typically 30 msec or so), if a successful connection is made, a rich
       PV object with will be stored internally for later use.  Use
       connection_timeout to control how long to wait before declaring that
       the PV cannot be connected (mispelled name, inactive IOC, improper
       network configuration)
    3. Since some PVs can take a long time to process (perhaps physically
       moving a motor or telling a detector to collect and not return until
       done), it is impossible to tell what a "reasonable" timeout for a put
       should be.

    Examples
    --------
    to put a value to a PV and return as soon as possible:

    >>> caput('xx.VAL', 3.0)

    to wait for processing to finish, use 'wait=True':

    >>> caput('xx.VAL', 3.0, wait=True)
    """
    start_time = time.time()
    thispv = get_pv(pvname, timeout=connection_timeout, connect=True)
    out = None
    if thispv.connected:
        timeout -= time.time() - start_time
        out = thispv.put(value, wait=wait, timeout=timeout)
    return out


def caget(pvname, as_string=False, count=None, as_numpy=True,
          use_monitor=True, timeout=5.0, connection_timeout=5.0):
    """get the current value to an epics Process Variable (PV).

    Parameters
    ----------
     pvname : str
        name of PV
     as_string : bool, optional
        whether to get the string representation [False]
     count : int or None
        maximum number of elements for waveform data [None]
     as_numpy : bool, optional
        whether to return waveform data as a numpy array [True]
     use_monitor : bool, optional
        whether to use the value cached by the monitor [True]
     timeout : float
        maximum time (in seconds) to wait for the processing to complete [60]
     connection_timeout :float
        maximum time (in seconds) to wait for connection [5]

    Returns
    -------
    data : object
       value of PV value on success, or ``None`` on a failure.

    Notes
    -----
    1. If not already connected, the PV will first attempt to connect to
       the networked variable. As the initial connection can take some time
       (typically 30 msec or so), if a successful connection is made, a rich
       PV object with will be stored internally for later use.
    2. `as_string=True` will return a string: float values will formatted
       according to the PVs precision, enum values will return the approriate
       enum string, etc.
    3. `use_monitor=True` will return the most recently cached value from the
       internal monitor placed on the PV. This will be the latest value unless
       the value is changing very rapidly. `use_monitor=False` will ignore the
       cached value and ask for an explicit value.  Of course, if a PV is
       changing rapidly enough for that to make a difference, it may also
       change between getting the value and downstream code using it.

    Examples
    --------
    to get the value of a PV:

    >>> x = caget('xx.VAL')

    to get the character string representation (formatted double, enum string, etc):

    >>> x = caget('xx.VAL', as_string=True)

    to get a truncated amount of data from an array, you can specify the count:

    >>> x = caget('MyArray.VAL', count=1000)
    """
    start_time = time.time()
    thispv = get_pv(pvname, timeout=connection_timeout, connect=True)
    val = None
    if thispv.connected:
        if as_string:
            thispv.get_ctrlvars()
        timeout -= time.time() - start_time
        val = thispv.get(count=count, timeout=timeout,
                         use_monitor=use_monitor,
                         as_string=as_string,
                         as_numpy=as_numpy)
        poll()
    return val

def cainfo(pvname, print_out=True, timeout=5.0, connection_timeout=5.0):
    """
    return printable information about PV

    Parameters
    ----------
     pvname  : str
         name of PV
     print_out : bool, optional
         whether to print the info to standard out [True]
     timeout : float
         maximum time (in seconds) to wait for the processing to complete [60]
     connection_timeout : float
         maximum time (in seconds) to wait for connection [5]

    Returns
    -------
    ``None`` or string :
         if `print_out=False`, the text is returned, but not printed.

    Notes
    -----
    1. If not already connected, the PV will first attempt to connect to
       the networked variable. As the initial connection can take some time
       (typically 30 msec or so), if a successful connection is made, a rich
       PV object with will be stored internally for later use.
    2. All time in seconds.


    Examples
    --------
    to print a status report for the PV:

    >>>cainfo('xx.VAL')

    to get the multiline text, use

    >>>txt = cainfo('xx.VAL', print_out=False)

    """
    start_time = time.time()
    thispv = get_pv(pvname, timeout=connection_timeout, connect=True)
    out = None
    if thispv.connected:
        conn_time = time.time() - start_time
        thispv.get(timeout=timeout-conn_time)
        get_time = time.time() - start_time
        thispv.get_ctrlvars(timeout=timeout-get_time)
        if print_out:
            ca.write(thispv.info)
        else:
            out = thispv.info
    return out

def camonitor_clear(pvname):
    """clear monitors on a PV

    Parameters
    ----------
     pvname  : str
         name of PV

    """
    if pvname in _PVmonitors_:
        _PVmonitors_[pvname].remove_callback(index=-999)
        _PVmonitors_.pop(pvname)

def camonitor(pvname, writer=None, callback=None, connection_timeout=5.0,
                  monitor_delta=None):
    """camonitor(pvname, writer=None, callback=None, connection_timeout=5)

    sets a monitor on a PV.

    Parameters
    ----------
     pvname : str
         name of PV
     writer : callable or ``None``
         function used to send monitor messages [None]
     callback :  callback or ``None``
         custom callback function [None]
     connection_timeout : float
         maximum time (in seconds) to wait for connection [5]
     monitor_delta : float or ``None``
         minimum change in value to monitor [None]

    Notes
    -----
    1. To write the result to a file, provide the `writer` option, such as a write
       method to an open file or some other callable that accepts a string.
    2. To completely control where the output goes, provide a `callback`
       method and you can do whatever you'd like with them.  This callback will be
       sent keyword arguments for pvname, value, char_value, and more. Use `**kws`!
    3. `monitor_delta` will set a value below which changes in the PV value
       will not generate an event.  This will try to set the `MDEL` field for
       a PV, limiting events from being sent from the PV server.  If MDEL
       cannot be set (perhaps due to write-access), this will be simulated in the
       client code. The default `None` will leave the current value unchanged.


    Examples
    --------
    to write a message with the latest value for that PV each time the value changes and
    when ca.poll() is called.

    >>>camonitor('xx.VAL')
    """

    if writer is None:
        writer = ca.write
    if callback is None:
        def callback(pvname=None, value=None, char_value=None, **kwds):
            "generic monitor callback"
            if char_value is None:
                char_value = repr(value)
            writer(f"{pvname:.32s} {pv.fmt_time()} {char_value}")

    thispv = get_pv(pvname, timeout=connection_timeout, connect=True,
                    monitor_delta=monitor_delta)
    if thispv.connected:
        thispv.get()
        thispv.add_callback(callback, index=-999, with_ctrlvars=True)
        _PVmonitors_[pvname] = thispv

def caget_many(pvlist, as_string=False, as_numpy=True, count=None,
               timeout=1.0, connection_timeout=5.0, conn_timeout=None):
    """
    get values for a list of PVs, working as fast as possible

    Parameters
    ----------
     pvlist : list of strings
         list of pv names to fetch
     as_string :  bool
         whether to get values as strings [False]
     as_numpy : bool
         whether to get values as numpy arrys [True]
     count : int or ``None``
         maximum number of elements to get [None]
     timeout :  float
         maximum time (in seconds) to wait for *each* get()  [1.0]
     connection_timeout : float
         maximum time (in seconds) to wait for *all* pvs to connect [5.0]
     conn_timeout :float
         back-compat alias or connection_timeout

    Returns
    -------
    list of values, with `None` signifying 'not connected' or 'timed out'.

    Notes
    -----
    this does not cache PV objects.

    """
    chids, connected, out = [], [], []
    for name in pvlist:
        chids.append(ca.create_channel(name, auto_cb=False, connect=False))

    all_connected = False
    if conn_timeout is not None:
        connection_timeout = conn_timeout
    expire_time = time.time() + connection_timeout
    while (not all_connected and (time.time() < expire_time)):
        connected = [dbr.CS_CONN==ca.state(chid) for chid in chids]
        all_connected = all(connected)
        poll()

    for (chid, conn) in zip(chids, connected):
        if conn:
            ca.get(chid, count=count, as_string=as_string, as_numpy=as_numpy,
                   wait=False)

    poll()
    for (chid, conn) in zip(chids, connected):
        val = None
        if conn:
            val = ca.get_complete(chid, count=count, as_string=as_string,
                                  as_numpy=as_numpy, timeout=timeout)
        out.append(val)
    return out

def caput_many(pvlist, values, wait=False, connection_timeout=5.0,
               put_timeout=60):
    """
    put values to a list of PVs, as quickly as possible

    Parameters
    ----------
     pvlist  : list or iterable
         list of PV names
     values : list or iterable
         list of values for corresponding PVs
     wait :   bool or string
         whether to wait for puts (see notes) [False]
     put_timeout : float
         maximum time (in seconds) to wait for the put to complete [60]
     connection_timeout :float
         maximum time (in seconds) to wait for connection [5]


    Returns
    -------
     a list of ints or `None`, with values of 1 if the put was successful,
     or a negative number if the put failed (say, the timeout was exceeded),
     or `None` if the connection failed.

    Notes
    -----
    1. This does not maintain the PV objects.
    2. With `wait='each'`, *each* put operation will block until it is
       complete or until the put_timeout duration expires.
       With `wait='all'`, this method will block until *all* put
       operations are complete, or until the put_timeout expires.
       This `wait` only applies to the put timeout, not the
       connection timeout.

    """
    if len(pvlist) != len(values):
        raise ValueError("List of PV names must be equal to list of values.")
    kwargs = {'auto_monitor': False, 'timeout': connection_timeout}
    pvs = [get_pv(name, **kwargs) for name in pvlist]

    wait_all = wait=='all'
    put_kws = {'wait': (wait=='each'), 'timeout': put_timeout,
               'use_complete': wait_all}
    put_ret = []
    for pvo, val in zip(pvs, values):
        put_ret.append(pvo.put(val, **put_kws))
    out = None
    if wait_all:
        start_time = time.time()
        while not all(((pv.connected and pv.put_complete) for pv in pvs)):
            ca.poll()
            elapsed_time = time.time() - start_time
            if elapsed_time > put_timeout:
                break
        out = [1 if (pv.connected and pv.put_complete) else -1 for pv in pvs]
    else:
        out = [val if val == 1 else -1 for val in put_ret]
    return out
